import pandas as pd
import sklearn.neighbors
import scipy.sparse as sp
import networkx as nx
import numpy as np
import torch
import argparse

def parameter_setting():

    parser = argparse.ArgumentParser(description='stClinic for predicting clinically-related niches')
    parser.add_argument('--input_dir',   '-IP', type = str, default = '/home/zuocm/Share_data/xiajunjie/stClinic/Datasets/DLPFC/',  help='data directory')
    parser.add_argument('--out_dir',   '-OP', type = str, default = '/home/zuocm/Share_data/xiajunjie/stClinic/Datasets/DLPFC/stClinic/',  help='data directory')
    parser.add_argument('--rad_cutoff',   '-RC', type = int, default = 150,    help='The radius value used to construct intra-edge (i.e., spatial nearest neighbors) for each slice')
    parser.add_argument('--k_cutoff',   '-KC', type = int, default = 6,    help='The number of spatial neighbors used to construct intra-edge')
    parser.add_argument('--k',   '-K', type = int, default = 5,    help='The number of mutual nearest neighbors used to construct inter-edges across slices')
    parser.add_argument('--n_top_genes',   '-NTG', type = int, default = 5000,    help='The number of highly variable genes selected for each slice')
    parser.add_argument('--n_centroids',   '-NCS', type = int, default = 7,    help='The number of components of the GMM')
    parser.add_argument('--lr_integration',   '-LRI', type = float, default = 0.0005,    help='The learning rate used by stClinic when extracting batch-corrected features in slices')
    parser.add_argument('--lr_prediction',   '-LRP', type = float, default = 0.05,    help='The learning rate used by supervised stClinic')


def Cal_Spatial_Net(adata, rad_cutoff=None, k_cutoff=None,
                    max_neigh=50, model='Radius', verbose=True):
    """\
    Construct the spatial neighbor networks.

    Parameters
    ----------
    adata
        AnnData object of scanpy package.
    rad_cutoff
        radius cutoff when model='Radius'
    k_cutoff
        The number of nearest neighbors when model='KNN'
    model
        The network construction model. When model=='Radius', the spot is connected to spots whose distance is less than rad_cutoff. When model=='KNN', the spot is connected to its first k_cutoff nearest neighbors.

    Returns
    -------
    The spatial networks are saved in adata.uns['Spatial_Net']
    """

    assert (model in ['Radius', 'KNN'])
    if verbose:
        print('------Calculating intra-spatial graph...')
    coor = pd.DataFrame(adata.obsm['spatial'])
    coor.index = adata.obs.index
    coor.columns = ['imagerow', 'imagecol']

    nbrs = sklearn.neighbors.NearestNeighbors(
        n_neighbors=max_neigh + 1, algorithm='ball_tree').fit(coor)
    distances, indices = nbrs.kneighbors(coor)
    if model == 'KNN':
        indices = indices[:, 1:k_cutoff + 1]
        distances = distances[:, 1:k_cutoff + 1]
    if model == 'Radius':
        indices = indices[:, 1:]
        distances = distances[:, 1:]

    KNN_list = []
    for it in range(indices.shape[0]):
        KNN_list.append(pd.DataFrame(zip([it] * indices.shape[1], indices[it, :], distances[it, :])))
    KNN_df = pd.concat(KNN_list)
    KNN_df.columns = ['Cell1', 'Cell2', 'Distance']

    Spatial_Net = KNN_df.copy()
    if model == 'Radius':
        Spatial_Net = KNN_df.loc[KNN_df['Distance'] < rad_cutoff,]
    id_cell_trans = dict(zip(range(coor.shape[0]), np.array(coor.index), ))
    Spatial_Net['Cell1'] = Spatial_Net['Cell1'].map(id_cell_trans)
    Spatial_Net['Cell2'] = Spatial_Net['Cell2'].map(id_cell_trans)
    # self_loops = pd.DataFrame(zip(Spatial_Net['Cell1'].unique(), Spatial_Net['Cell1'].unique(),
    #                  [0] * len((Spatial_Net['Cell1'].unique())))) ###add self loops
    # self_loops.columns = ['Cell1', 'Cell2', 'Distance']
    # Spatial_Net = pd.concat([Spatial_Net, self_loops], axis=0)

    if verbose:
        print('The graph contains %d edges, %d cells.' % (Spatial_Net.shape[0], adata.n_obs))
        print('%.4f neighbors per spot on average.' % (Spatial_Net.shape[0] / adata.n_obs))
    adata.uns['Spatial_Net'] = Spatial_Net

    #########
    X = pd.DataFrame(adata.X.toarray()[:, ], index=adata.obs.index, columns=adata.var.index)
    cells = np.array(X.index)
    cells_id_tran = dict(zip(cells, range(cells.shape[0])))
    if 'Spatial_Net' not in adata.uns.keys():
        raise ValueError("Spatial_Net is not existed! Run Cal_Spatial_Net first!")
        
    Spatial_Net = adata.uns['Spatial_Net']
    G_df = Spatial_Net.copy()
    G_df['Cell1'] = G_df['Cell1'].map(cells_id_tran)
    G_df['Cell2'] = G_df['Cell2'].map(cells_id_tran)
    G = sp.coo_matrix((np.ones(G_df.shape[0]), (G_df['Cell1'], G_df['Cell2'])), shape=(adata.n_obs, adata.n_obs))
    G = G + sp.eye(G.shape[0])  # self-loop
    adata.uns['adj'] = G



def mclust_R(adata, num_cluster, modelNames='EEE', used_obsm='VGAEX', random_seed=666):
    """\
    Clustering using the mclust algorithm.
    The parameters are the same as those in the R package mclust.
    """

    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("mclust")

    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']

    mclust_res = np.array(rmclust(np.array(adata.obsm[used_obsm]), num_cluster, modelNames)[-2])

    adata.obs['mclust'] = mclust_res
    adata.obs['mclust'] = adata.obs['mclust'].astype('int')
    adata.obs['mclust'] = adata.obs['mclust'].astype('category')
    return adata



def match_cluster_labels(true_labels,est_labels):
    true_labels_arr = np.array(list(true_labels))
    est_labels_arr = np.array(list(est_labels))
    org_cat = list(np.sort(list(pd.unique(true_labels))))
    est_cat = list(np.sort(list(pd.unique(est_labels))))
    B = nx.Graph()
    B.add_nodes_from([i+1 for i in range(len(org_cat))], bipartite=0)
    B.add_nodes_from([-j-1 for j in range(len(est_cat))], bipartite=1)
    for i in range(len(org_cat)):
        for j in range(len(est_cat)):
            weight = np.sum((true_labels_arr==org_cat[i])* (est_labels_arr==est_cat[j]))
            B.add_edge(i+1,-j-1, weight=-weight)
    match = nx.algorithms.bipartite.matching.minimum_weight_full_matching(B)
#     match = minimum_weight_full_matching(B)
    if len(org_cat)>=len(est_cat):
        return np.array([match[-est_cat.index(c)-1]-1 for c in est_labels_arr])
    else:
        unmatched = [c for c in est_cat if not (-est_cat.index(c)-1) in match.keys()]
        l = []
        for c in est_labels_arr:
            if (-est_cat.index(c)-1) in match: 
                l.append(match[-est_cat.index(c)-1]-1)
            else:
                l.append(len(org_cat)+unmatched.index(c))
        return np.array(l)  



def count_params(model):
    all_params = sum(p.numel() for p in model.parameters())
    trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    non_trainable_params = sum(p.numel() for p in model.parameters() if not p.requires_grad)

    print(f'Total params: {all_params}\nTrainable params: {trainable_params}\nNon-trainable params: {non_trainable_params}')



from sklearn.preprocessing import scale
from scipy.sparse.linalg import eigsh
def estimate_k(data):
    """
    Estimate number of groups k:
        based on random matrix theory (RTM), borrowed from SC3
        input data is (p,n) matrix, p is feature, n is sample
    """
    p, n = data.shape

    x = scale(data, with_mean=False)
    muTW = (np.sqrt(n-1) + np.sqrt(p)) ** 2
    sigmaTW = (np.sqrt(n-1) + np.sqrt(p)) * (1/np.sqrt(n-1) + 1/np.sqrt(p)) ** (1/3)
    sigmaHatNaive = x.T.dot(x)

    bd = np.sqrt(p) * sigmaTW + muTW
    evals, _ = eigsh(sigmaHatNaive)

    k = 0
    for i in range(len(evals)):
        if evals[i] > bd:
            k += 1
    return k



from sklearn.metrics import silhouette_score as s_score
def F1score(adata):
    """
    Compute silhouette coefficient:
    F1score = 2*(1-silh'batch)*silh'cluster / [silh'cluster + (1-silh'batch)]
    silh'batch = (1+silh_batch)/2
    silh'cluster = (1+silh_cluster)/2
    """
    s_cluster = s_score(adata.obsm['X_umap'], adata.obs['mclust'])
    s_batch = s_score(adata.obsm['X_umap'], adata.obs['batch_name'])

    s_cluster = (1+s_cluster)/2
    s_batch = (1+s_batch)/2

    F1score = 2*(1-s_batch)*s_cluster / (s_cluster + (1-s_batch))

    return F1score



def LISI_score(adata, random_seed=666):
    """
    Compute cell-type / integration local inverse simpson'index
    """

    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("lisi")
    rlisi = robjects.r['compute_lisi']

    import rpy2.robjects.numpy2ri, rpy2.robjects.pandas2ri
    rpy2.robjects.numpy2ri.activate()
    rpy2.robjects.pandas2ri.activate()

    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)

    robjects.globalenv['X'] = np.array(adata.obsm['X_umap'])
    meta_data = adata.obs[['Ground Truth', 'batch_name']]
    meta_data.columns = ['label1','label2']
    robjects.globalenv['meta_data'] = meta_data
    robjects.globalenv['cnames'] = robjects.r['c']('label1', 'label2')

    lisi_res = np.array(
                        rlisi(robjects.globalenv['X'], robjects.globalenv['meta_data'], robjects.globalenv['cnames'])
                        )

    clisi = np.mean(lisi_res, axis=1)[0]
    ilisi = np.mean(lisi_res, axis=1)[1]

    return dict(cLISI = clisi, iLISI = ilisi, LISI=lisi_res)



from lifelines.utils import concordance_index
from lifelines.statistics import logrank_test
def CIndex_lifeline(hazards, labels, survtime_all):
    return(concordance_index(survtime_all, -hazards, labels))



def cox_log_rank(hazardsdata, labels, survtime_all):
    median = np.median(hazardsdata)
    hazards_dichotomize = np.zeros([len(hazardsdata)], dtype=int)
    hazards_dichotomize[hazardsdata > median] = 1
    idx = hazards_dichotomize == 0
    T1 = survtime_all[idx]
    T2 = survtime_all[~idx]
    E1 = labels[idx]
    E2 = labels[~idx]
    results = logrank_test(T1, T2, event_observed_A=E1, event_observed_B=E2)
    pvalue_pred = results.p_value
    return(pvalue_pred)



def accuracy_cox(hazardsdata, labels):
    # This accuracy is based on estimated survival events against true survival events
    median = np.median(hazardsdata)
    hazards_dichotomize = np.zeros([len(hazardsdata)], dtype=int)
    hazards_dichotomize[hazardsdata > median] = 1
    correct = np.sum(hazards_dichotomize == labels)
    return correct / len(labels)



from lifelines import KaplanMeierFitter
from lifelines import CoxPHFitter
def KM_plot(hazardsdata, labels, survtime_all, output_dir):
    median = np.median(hazardsdata)
    hazards_dichotomize = np.zeros([len(hazardsdata)], dtype=int)
    hazards_dichotomize[hazardsdata > median] = 1
    idx = (hazards_dichotomize==1)

    T = survtime_all
    E = labels

    import matplotlib.pyplot as plt
    plt.figure(figsize=(8, 8))
    ax = plt.subplot(111)

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_linewidth(1)
    ax.spines['bottom'].set_linewidth(1)

    kmf_high = KaplanMeierFitter()
    ax = kmf_high.fit(T[idx], E[idx], label='high risk (n={})'.format(sum(idx))).plot_survival_function(linewidth=3, show_censors=True, censor_styles={'ms':9, 'marker':'+'}, ax=ax, ci_show=False, color='#FF4B68')

    kmf_low = KaplanMeierFitter()
    ax = kmf_low.fit(T[~idx], E[~idx], label='low risk (n={})'.format(len(idx)-sum(idx))).plot_survival_function(linewidth=3, show_censors=True, censor_styles={'ms':9, 'marker':'+'}, ax=ax, ci_show=False, color='#118DF0')

    legend = plt.legend(loc="upper right",fontsize=10)
    legend.get_frame().set_facecolor('none')
    legend.get_frame().set_linewidth(0.0)
    plt.xlabel('Time (days)',fontsize=12)
    plt.ylim([-0.01, 1.01])
    plt.ylabel('Population at risk (%)',fontsize=12)
    plt.xticks(fontsize=12)
    plt.yticks(fontsize=12)

    logrank_stat = logrank_test(T[idx], T[~idx], E[idx], E[~idx], alpha=.99).p_value
    plt.text(7.5,0.05,s='p={:.2e}'.format(logrank_stat),fontsize=10)

    cph = CoxPHFitter()
    df = pd.DataFrame({'Surv_time': T, 'Event':T, 'High_risk':hazards_dichotomize})
    cph.fit(df, 'Surv_time', 'Event')
    hr_value = cph.summary.loc['High_risk', 'exp(coef)']
    ci_lower = cph.summary.loc['High_risk', 'exp(coef) lower 95%']
    ci_upper = cph.summary.loc['High_risk', 'exp(coef) upper 95%']

    plt.text(7.5,0.07, s=f'HR: {hr_value:.3f} ({ci_lower:.3f}, {ci_upper:.3f})', fontsize=10)

    plt.savefig(output_dir + 'Stat_AttPred_UMAP_KM_plot.jpg', dpi=500)



from sklearn.metrics import confusion_matrix
def CM_plot(pred_data, labels, output_dir):
    cm = confusion_matrix(labels, pred_data)

    import matplotlib.pyplot as plt
    import seaborn as sns
    plt.figure(figsize=(6,6))
    sns.heatmap(cm, annot=True, cmap='OrRd', fmt='g', annot_kws={'fontsize': 12})
    plt.xlabel('Ground-truth label', fontdict={'size':12})
    plt.ylabel('Predicted label', fontdict={'size':12})

    plt.savefig(output_dir + 'Stat_AttPred_UMAP_CM_plot.jpg', dpi=500)



from sklearn.metrics import roc_curve, auc
def ROC_plot(pred_data, labels, output_dir):
    fpr, tpr, thresholds = roc_curve(labels, pred_data)
    auc_score = auc(fpr,tpr)

    import matplotlib.pyplot as plt
    plt.figure(figsize=(6,6))
    plt.plot(fpr, tpr, label=f'AUC: {auc_score:.3f}')
    plt.plot([0,1],[0,1],linestyle='--',color='grey')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.title('ROC curve')
    plt.xlabel('False Positive Rate',fontdict={'size':12})
    plt.ylabel('True Positive Rate',fontdict={'size':12})
    plt.legend(fontsize=12,loc='lower right')

    plt.savefig(output_dir + 'Stat_AttPred_UMAP_ROC_plot.jpg', dpi=500)



from sklearn.metrics import pairwise_distances
def knn_smoothing(latent, k, mat):
    dist = pairwise_distances(latent)
    row = []
    col = []
    sorted_knn = dist.argsort(axis=1)
    for idx in list(range(np.shape(dist)[0])):
        col.extend(sorted_knn[idx, : k].tolist())
        row.extend([idx] * k)

    res = np.zeros((mat.shape[0], mat.shape[1]))
    for i in range(len(col)):
        res[row[i]] += mat[col[i]]

    return res



from typing import List, Mapping, Optional, Union
import faiss
from sklearn.metrics.pairwise import euclidean_distances
from anndata import AnnData
def spatial_match(embds:List[torch.Tensor],
                  reorder:Optional[bool]=True,
                  smooth:Optional[bool]=True,
                  smooth_range:Optional[int]=20,
                  scale_coord:Optional[bool]=True,
                  adatas:Optional[List[AnnData]]=None,
                  return_euclid:Optional[bool]=False,
                  verbose:Optional[bool]=False,
                  get_null_distri:Optional[bool]=False
    )-> List[Union[np.ndarray,torch.Tensor]]:
    r"""
    Use embedding to match cells from different batches based on cosine similarity
    
    Parameters
    ----------
    embds
        list of embeddings
    reorder
        if reorder embedding by cell numbers
    smooth
        if smooth the mapping by Euclid distance
    smooth_range
        use how many candidates to do smooth
    scale_coord
        if scale the coordinate to [0,1]
    adatas
        list of adata object
    verbose
        if print log
    get_null_distri
        if get null distribution of cosine similarity
    
    Note
    ----------
    Automatically use larger dataset as source
    
    Return
    ----------
    Best matching, Top n matching and cosine similarity matrix of top n  
    """
    if reorder and embds[0].shape[0] < embds[1].shape[0]:
        embd0 = embds[1]
        embd1 = embds[0]
        adatas = adatas[::-1] if adatas is not None else None
    else:
        embd0 = embds[0]
        embd1 = embds[1]
        
    if get_null_distri:
        embd0 = torch.tensor(embd0)
        embd1 = torch.tensor(embd1)
        sample1_index = torch.randint(0, embd0.shape[0], (1000,))
        sample2_index = torch.randint(0, embd1.shape[0], (1000,))
        cos = torch.nn.CosineSimilarity(dim=1)
        null_distri = cos(embd0[sample1_index], embd1[sample2_index])

    index = faiss.index_factory(embd1.shape[1], "Flat", faiss.METRIC_INNER_PRODUCT)
    embd0_np = embd0.detach().cpu().numpy() if torch.is_tensor(embd0) else embd0
    embd1_np = embd1.detach().cpu().numpy() if torch.is_tensor(embd1) else embd1
    embd0_np = embd0_np.copy().astype('float32')
    embd1_np = embd1_np.copy().astype('float32')
    faiss.normalize_L2(embd0_np)
    faiss.normalize_L2(embd1_np)
    index.add(embd0_np)
    similarity, order = index.search(embd1_np, smooth_range)
    best = []
    if smooth and adatas != None:
        if verbose:
            print('Smoothing mapping, make sure object is in same direction')
        if scale_coord:
            # scale spatial coordinate of every adata to [0,1]
            adata1_coord = adatas[0].obsm['spatial'].copy()
            adata2_coord = adatas[1].obsm['spatial'].copy()
            for i in range(2):
                    adata1_coord[:,i] = (adata1_coord[:,i]-np.min(adata1_coord[:,i]))/(np.max(adata1_coord[:,i])-np.min(adata1_coord[:,i]))
                    adata2_coord[:,i] = (adata2_coord[:,i]-np.min(adata2_coord[:,i]))/(np.max(adata2_coord[:,i])-np.min(adata2_coord[:,i]))
        dis_list = []
        for query in range(embd1_np.shape[0]):
            ref_list = order[query, :smooth_range]
            dis = euclidean_distances(adata2_coord[query,:].reshape(1, -1),
                                      adata1_coord[ref_list,:])
            dis_list.append(dis)
            best.append(ref_list[np.argmin(dis)])
    else:
        best = order[:,0]

    if return_euclid and smooth and adatas != None:
        dis_array = np.squeeze(np.array(dis_list))
        if get_null_distri:
            return np.array(best), order, similarity, dis_array, null_distri
        else:
            return np.array(best), order, similarity, dis_array
    else:
        return np.array(best), order, similarity



class EarlyStopping:
    """
    Early stops the training if loss doesn't improve after a given patience.
    """
    def __init__(self, patience=10, verbose=False, checkpoint_file=''):
        """
        Parameters
        ----------
        patience 
            How long to wait after last time loss improved. Default: 10
        verbose
            If True, prints a message for each loss improvement. Default: False
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.loss_min = np.Inf
        self.checkpoint_file = checkpoint_file

    def __call__(self, loss, model):
        if np.isnan(loss):
            self.early_stop = True
        score = -loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(loss, model)
        elif score <= self.best_score:
            self.counter += 1
            if self.verbose:
                print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
                model.load_model(self.checkpoint_file)
        else:
            self.best_score = score
            self.save_checkpoint(loss, model)
            self.counter = 0

    def save_checkpoint(self, loss, model):
        '''
        Saves model when loss decrease.
        '''
        if self.verbose:
            print(f'Loss decreased ({self.loss_min:.6f} --> {loss:.6f}).  Saving model ...')
        if self.checkpoint_file:
            torch.save(model.state_dict(), self.checkpoint_file)
        self.loss_min = loss


import numpy as np
import random
import os
import torch
import torch.nn as nn
import torch.backends.cudnn as cudnn
import math
import re
import torch.nn.functional as F
import torch_geometric.transforms as T
import torch.optim as optim

from tqdm import tqdm
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from sklearn.metrics.pairwise import pairwise_distances
from torch_geometric.nn import InnerProductDecoder
from torch_geometric.utils import (negative_sampling, remove_self_loops, add_self_loops)
from tqdm.autonotebook import trange
from torch_geometric.loader import DenseDataLoader
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader

from .Utilities import *
from .gat_conv import GATConv

cudnn.deterministic = True
cudnn.benchmark = True


class DSBatchNorm(nn.Module):
    """
    Domain-specific Batch Normalization for each slice
    """
    def __init__(self, num_features, n_domains, eps=1e-5, momentum=0.1):
        super().__init__()
        self.n_domains = n_domains
        self.num_features = num_features
        self.bns = nn.ModuleList([nn.BatchNorm1d(num_features, eps=eps, momentum=momentum) for i in range(n_domains)])
        
    def reset_running_stats(self):
        for bn in self.bns:
            bn.reset_running_stats()
            
    def reset_parameters(self):
        for bn in self.bns:
            bn.reset_parameters()
            
    def _check_input_dim(self, input):
        raise NotImplementedError
            
    def forward(self, x, y):
        out = torch.zeros(x.size(0), self.num_features, device=x.device)
        for i in range(self.n_domains):
            indices = np.where(y.cpu().numpy()==i)[0]

            if len(indices) > 1:
                out[indices] = self.bns[i](x[indices])
            elif len(indices) == 1:
                out[indices] = x[indices]

        return out



class stClinic(nn.Module):
    def __init__(self, hidden_dims_integ, n_domains, n_centroids):
        super(stClinic, self).__init__()

        # Set network structure
        [in_dim, num_hidden, out_dim] = hidden_dims_integ

        self.fc1      = nn.Linear(in_dim, in_dim)
        self.norm     = nn.BatchNorm1d(in_dim)
        self.dropout1 = nn.Dropout(0.2)

        self.conv1 = GATConv(in_dim, num_hidden, heads=3, concat=False,
                             dropout=0.2, add_self_loops=False, bias=False)

        self.conv_mu = GATConv(num_hidden, out_dim, heads=3, concat=False,
                               dropout=0.2, add_self_loops=False, bias=False)
        self.conv_log_var = GATConv(num_hidden, out_dim, heads=3, concat=False,
                                    dropout=0.2, add_self_loops=False, bias=True)

        self.fc2      = nn.Linear(out_dim, in_dim)
        self.dsbnorm  = DSBatchNorm(in_dim, n_domains)
        self.dropout2 = nn.Dropout(0.2)
        self.fc3 = nn.Linear(in_dim, in_dim)

        self.graph_decoder = InnerProductDecoder()

        # init c parameters
        self.n_centroids = n_centroids

        self.pi    = nn.Parameter(torch.ones(n_centroids)/n_centroids) # p(c)
        self.mu_c  = nn.Parameter(torch.zeros(out_dim, n_centroids)) # mu
        self.var_c = nn.Parameter(torch.ones(out_dim, n_centroids)) # sigma^2

    def reparameterize(self, mu, log_var):
        if self.training:
            return mu + torch.randn_like(log_var) * torch.exp(log_var.mul(0.5))
        else:
            return mu

    def get_gamma(self, z):
        """
        Inference c from z

        gamma equals q(c|x)

        q(c|x) = p(c|z) = p(c)*p(z|c)/p(z)
        """
        n_centroids = self.n_centroids

        N     = z.size(0)
        z     = z.unsqueeze(2).expand(z.size(0), z.size(1), n_centroids)
        pi    = self.pi.repeat(N, 1) # N*K
        mu_c  = self.mu_c.repeat(N, 1, 1) # N*D*K
        var_c = self.var_c.repeat(N, 1, 1) + 1e-8 # N*D*K

        # p(c, z) = p(c)*p(z|c)
        p_c_z = torch.exp(torch.log(pi) - torch.sum(0.5 * torch.log(2 * math.pi * var_c) + (z - mu_c)**2/(2 * var_c), dim=1)) + 1e-10
        gamma = p_c_z / torch.sum(p_c_z, dim=1, keepdim=True)

        return gamma, mu_c, var_c, pi

    def init_gmm_params(self, features, edge_index, adj_mat, type='GMM', seed=666, start_stage=False):
        """
        Initialize stClinic integration model with GMM model parameters / Prune edges
        """
        if type == 'GMM':

            gmm = GaussianMixture(n_components=self.n_centroids, covariance_type='diag', random_state=seed)

            h0 = self.dropout1( F.elu( self.norm( self.fc1(features) ) ) )

            h1 = F.elu(self.conv1(h0, edge_index, attention=True))

            mu = self.conv_mu(h1, edge_index, attention=True)
            log_var = self.conv_log_var(h1, edge_index, attention=True)

            z = mu
            z = z.cpu().detach().numpy()

            gmm.fit(z)
            labels = gmm.predict(z)

            if start_stage:
                print('Selected initial type: GMM')

                '''
                Variance Reduction
                '''
                sampled_z = []
                for c in range(self.n_centroids):
                    z_c = z[labels==c,:]
                    mean_c = gmm.means_[c,:]
                    sampled_z_c = z_c[np.argsort(np.sum((z_c - mean_c)**2, axis=1)) <= int(z_c.shape[0]/2),:]
                    sampled_z.append(sampled_z_c)
                sampled_z = np.concatenate(sampled_z, axis=0)

                sampled_gmm = GaussianMixture(n_components=self.n_centroids, covariance_type='diag', random_state=seed)
                sampled_gmm.fit(sampled_z)

                self.mu_c.data.copy_(torch.from_numpy(sampled_gmm.means_.T.astype(np.float32)))
                self.var_c.data.copy_(torch.from_numpy(sampled_gmm.covariances_.T.astype(np.float32)))

            else:
                print('Prune edges...')

                prune_mat = torch.IntTensor(np.where(pairwise_distances(labels.reshape(-1,1), metric='hamming')==0, 1, 0)).cuda()
                pruned_adj = prune_mat.to(torch.int8) * adj_mat.to(torch.int8)

                return pruned_adj

        elif type == 'KMeans':

            h0 = self.dropout1( F.elu( self.norm( self.fc1(features) ) ) )

            h1 = F.elu(self.conv1(h0, edge_index, attention=True))

            mu = self.conv_mu(h1, edge_index, attention=True)
            log_var = self.conv_log_var(h1, edge_index, attention=True)

            z = mu

            kmeans = KMeans(n_clusters=self.n_centroids, random_state=seed).fit(z.cpu().detach().numpy())
            mean = kmeans.cluster_centers_
            cls_index = kmeans.labels_

            if start_stage:
                print('Selected initial type: KMeans')

                '''
                Variance Reduction
                '''
                sampled_z = []
                for c in range(self.n_centroids):
                    z_c = z.cpu().detach().numpy()[cls_index==c,:]
                    mean_c = mean[c,:]
                    sampled_z_c = z_c[np.argsort(np.sum((z_c - mean_c)**2, axis=1)) <= int(z_c.shape[0]/2),:]
                    sampled_z.append(sampled_z_c)
                sampled_z = np.concatenate(sampled_z, axis=0)

                sampled_kmeans = KMeans(n_clusters=self.n_centroids, random_state=seed).fit(sampled_z)
                sampled_mean = sampled_kmeans.cluster_centers_
                sampled_cls_index = sampled_kmeans.labels_

                sampled_var = []
                sampled_mean = torch.from_numpy(sampled_mean).cuda()
                sampled_z = torch.from_numpy(sampled_z).cuda()
                for c in range(self.n_centroids):
                    index = np.where(sampled_cls_index==c)
                    var_g = torch.sum((sampled_z[index[0],:]-sampled_mean[c,:])**2, dim=0, keepdim=True)/(len(index[0])-1)
                    sampled_var.append(var_g)
                sampled_var = torch.cat(sampled_var, dim=0)
                sampled_mean = sampled_mean.cpu().detach().numpy().astype(np.float32)
                sampled_var = sampled_var.cpu().detach().numpy().astype(np.float32)

                self.mu_c.data.copy_(torch.from_numpy(sampled_mean.transpose()))
                self.var_c.data.copy_(torch.from_numpy(sampled_var.transpose()))
            
            else:
                print('Prune edges...')

                prune_mat = torch.IntTensor(np.where(pairwise_distances(cls_index.reshape(-1,1), metric='hamming')==0, 1, 0)).cuda()
                pruned_adj = prune_mat.to(torch.int8) * adj_mat.to(torch.int8)

                return pruned_adj

    def forward(self, features, edge_index, y):

        h0 = self.dropout1( F.elu( self.norm( self.fc1(features) ) ) )

        h1 = F.elu(self.conv1(h0, edge_index, attention=True))

        mu      = self.conv_mu(h1, edge_index, attention=True)
        log_var = self.conv_log_var(h1, edge_index, attention=True)

        z  = self.reparameterize(mu, log_var)

        h2 = self.dropout2( F.elu( self.dsbnorm( self.fc2(z), y ) ) )

        h3 = self.fc3(h2)

        return z, mu, log_var, h3
    
    def graph_recon_loss(self, z, pos_edge_label_index):

        EPS = 1e-15

        neg_edge_index = None

        reG = self.graph_decoder(z, pos_edge_label_index, sigmoid=True)
        pos_loss = -torch.log(reG + EPS).mean()
        pos_edge_index, _ = remove_self_loops(pos_edge_label_index)
        pos_edge_index, _ = add_self_loops(pos_edge_index)
        if neg_edge_index is None:
            neg_edge_index = negative_sampling(pos_edge_index, z.size(0))
        neg_loss = -torch.log(1 - self.graph_decoder(z, neg_edge_index, sigmoid=True) + EPS).mean()
        graph_recon_loss = pos_loss + neg_loss

        return graph_recon_loss
    
    def KL_loss(self, z, mu, log_var):

        gamma, mu_c, var_c, pi = self.get_gamma(z)
        var_c += 1e-8
        n_c = pi.size(1)
        mu_expand = mu.unsqueeze(2).expand(mu.size(0), mu.size(1), n_c)
        logvar_expand = log_var.unsqueeze(2).expand(log_var.size(0), log_var.size(1), n_c)
        
        # log p(z|c)
        logpzc = -0.5*torch.sum(gamma*torch.sum(math.log(2*math.pi) + \
                                                torch.log(var_c) + \
                                                torch.exp(logvar_expand)/var_c + \
                                                (mu_expand-mu_c)**2/var_c, dim=1), dim=1)

        # log p(c)
        logpc = torch.sum(gamma*torch.log(pi), 1)

        # log q(z|x)
        qentropy = -0.5*torch.sum(1+log_var, 1)

        # log q(c|x)
        logqcx = torch.sum(gamma*torch.log(gamma), 1)

        # torch.mean or (1 / data.x.shape[0]) * torch.sum
        KL_loss = torch.mean( -logpzc - logpc + qentropy + logqcx )

        return KL_loss

    def load_model(self, path):
        """
        Load trained model parameters dictionary.
        Parameters
        ----------
        path
            file path that stores the model parameters
        """
        pretrained_dict = torch.load(path, map_location=lambda storage, loc: storage)
        model_dict = self.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if k in model_dict}
        model_dict.update(pretrained_dict) 
        self.load_state_dict(model_dict)



def train_stClinic_model(adata, lr=0.0005, n_centroids=7, MultiOmics_mode=False,
                         used_feat='X_seurat', hidden_dims_integ=[512, 10],
                         n_epochs_pre=300, n_epochs_tune=300+100, n_epochs_prune=50,
                         batch_name='batch_name', lambda_KL_stage1=0.5, lambda_KL_stage2=0.2,
                         gradient_clipping=5, weight_decay=0.00001, key_add='stClinic',
                         random_seed=666, device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):

    seed = random_seed
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)

    adj_mat  = torch.IntTensor(adata.uns['adj']).to(device)
    edgeList = adata.uns['edgeList']

    if MultiOmics_mode:

        data = Data(edge_index=torch.LongTensor(np.array([edgeList[0], edgeList[1]])),
                    prune_edge_index=torch.LongTensor(np.array([])),
                    x=torch.FloatTensor(adata.obsm[used_feat]),
                    y=torch.LongTensor(adata.obs[batch_name].cat.codes))

    else:

        data = Data(edge_index=torch.LongTensor(np.array([edgeList[0], edgeList[1]])),
                    prune_edge_index=torch.LongTensor(np.array([])),
                    x=torch.FloatTensor(adata.X.todense()),
                    y=torch.LongTensor(adata.obs[batch_name].cat.codes))

    # Transform
    transform  = T.RandomLinkSplit(num_val=0, num_test=0, is_undirected=True, add_negative_train_samples=False, split_labels=True)
    data, _, _ = transform(data)
    data       = data.to(device)

    model = stClinic(hidden_dims_integ=[data.x.shape[1], hidden_dims_integ[0], hidden_dims_integ[1]],
                     n_domains=len(adata.obs[batch_name].cat.categories),
                     n_centroids=n_centroids).to(device)

    print('Pretrain with unregularized stClinic (without GMM regularizer)...')
    model.pi.requires_grad, model.mu_c.requires_grad, model.var_c.requires_grad = False, False, False

    optimizer_stClinic = optim.Adam(filter(lambda p: p.requires_grad, model.parameters()), lr=lr, weight_decay=weight_decay)

    with trange(n_epochs_pre, total=n_epochs_pre, desc='Epochs') as tq:
        for epoch in tq:
            model.train()

            optimizer_stClinic.zero_grad()

            z, mu, log_var, reX = model(data.x, data.edge_index, data.y)

            features_recon_loss = F.mse_loss(data.x, reX)

            graph_recon_loss = model.graph_recon_loss(mu, data.pos_edge_label_index)

            KL_loss = 0

            epoch_loss = {'feat_recon_loss':features_recon_loss*10, 'graph_recon_loss':graph_recon_loss, 'KL_loss':KL_loss*0.0}

            sum(epoch_loss.values()).backward()
            nn.utils.clip_grad_norm_(filter(lambda p: p.requires_grad, model.parameters()), gradient_clipping)
            optimizer_stClinic.step()

            epoch_info = ','.join(['{}={:.3f}'.format(k, v) for k,v in epoch_loss.items()])
            tq.set_postfix_str(epoch_info) 

    model.eval()

    print('Initialize Gaussian Mixture Model...')
    model.init_gmm_params(data.x, data.edge_index, adj_mat, type='GMM', seed=random_seed, start_stage=True)

    print('Train with regularized stClinic (with GMM regularizer)...')
    model.pi.requires_grad, model.mu_c.requires_grad, model.var_c.requires_grad = True, True, True

    with trange(n_epochs_tune, total=n_epochs_tune, desc='Epochs') as tq:
        for epoch in tq:

            if epoch % n_epochs_prune == 0 and epoch >= 300:
                model.eval()
                pruned_adj = model.init_gmm_params(data.x, data.edge_index, adj_mat, type='GMM', seed=random_seed, start_stage=False)
                pruned_edgeList = np.nonzero(pruned_adj.cpu().detach().numpy())

                if MultiOmics_mode:

                    data = Data(edge_index=torch.LongTensor(np.array([pruned_edgeList[0], pruned_edgeList[1]])),
                                prune_edge_index=torch.LongTensor(np.array([])),
                                x=torch.FloatTensor(adata.obsm[used_feat]),
                                y=torch.LongTensor(adata.obs[batch_name].cat.codes))

                else:

                    data = Data(edge_index=torch.LongTensor(np.array([pruned_edgeList[0], pruned_edgeList[1]])),
                                prune_edge_index=torch.LongTensor(np.array([])),
                                x=torch.FloatTensor(adata.X.todense()),
                                y=torch.LongTensor(adata.obs[batch_name].cat.codes))

                # Transform
                transform  = T.RandomLinkSplit(num_val=0, num_test=0, is_undirected=True, add_negative_train_samples=False, split_labels=True)
                data, _, _ = transform(data)

                data = data.to(device)

            model.train()

            optimizer_stClinic.zero_grad()

            z, mu, log_var, reX = model(data.x, data.edge_index, data.y)

            # Compute features reconstruction loss  --default coefficient=10 
            features_recon_loss = F.mse_loss(data.x, reX)
            # print(f'MSE loss: {features_recon_loss}')

            # Compute graph reconstruction loss
            graph_recon_loss    = model.graph_recon_loss(mu, data.pos_edge_label_index)
            # print(f'BCE loss: {graph_recon_loss}')

            # Compute KL loss  --coefficient=0.0(like STAGATE)  --coefficient=0.5 (stage I)  --coefficient=0.2 (stage II)
            KL_loss = model.KL_loss(z, mu, log_var)
            # print(f'KL loss: {KL_loss}')

            if epoch < 300:
                epoch_loss = {'feat_recon_loss':features_recon_loss*10, 'graph_recon_loss':graph_recon_loss, 'KL_loss':KL_loss*lambda_KL_stage1}
            else:
                epoch_loss = {'feat_recon_loss':features_recon_loss*10, 'graph_recon_loss':graph_recon_loss, 'KL_loss':KL_loss*lambda_KL_stage2}

            sum(epoch_loss.values()).backward()
            nn.utils.clip_grad_norm_(filter(lambda p: p.requires_grad, model.parameters()), gradient_clipping)
            optimizer_stClinic.step()

            epoch_info = ','.join(['{}={:.3f}'.format(k, v) for k,v in epoch_loss.items()])
            tq.set_postfix_str(epoch_info) 

            # early_stopping(sum(epoch_loss.values()).cpu().detach().numpy(), model)
            # if early_stopping.early_stop:
            #     print('EarlyStopping: run {} epoch'.format(epoch+1))
            #     break

    model.eval()

    z, mu, _, _ = model(data.x, data.edge_index, data.y)
    adata.obsm[key_add] = mu.cpu().detach().numpy()
    gamma, _, _, _ = model.get_gamma(z)
    adata.obs['GMM_cluster'] = np.argmax(gamma.cpu().detach().numpy(), axis=1)
    adata.obs['GMM_cluster'] = adata.obs['GMM_cluster'].astype('category')

    return adata



def stClinic_Statistics_Measures(adata, sorted_batch):

    # Statistical representations of all slices
    All_cluster_repr = []
    for bid in sorted_batch:

        batch_obs = adata.obs[ adata.obs['batch_name'] == bid ]
        batch_repr = adata.obsm['X_umap'][ adata.obs['batch_name'] == bid, :]

        res_obs = adata.obs[ adata.obs['batch_name'] != bid ]

        # Statistical representations of each slice (Before transpose)
        batch_cluster_repr = []
        for cluster in np.unique(adata.obs['louvain']):
            if cluster in np.unique(batch_obs['louvain']):

                repr_init = batch_repr[ batch_obs['louvain'] == cluster, :]

                repr_mean = np.mean( repr_init, axis=0 )
                repr_var = np.mean(repr_init * repr_init, axis=0) - repr_mean * repr_mean

                repr_L2 = np.sum(repr_init * repr_init,axis=1)
                repr_max, repr_min = repr_init[np.argmax(repr_L2),:], repr_init[np.argmin(repr_L2),:]

                repr_p1 = np.array([repr_init.shape[0] / batch_repr.shape[0]])

                repr = np.concatenate([repr_mean, repr_var, repr_max, repr_min, repr_p1])

            else:
                repr = np.zeros(9)

            repr_p2 = np.array([res_obs[res_obs['louvain']==cluster].shape[0] / res_obs.shape[0]])

            repr = np.concatenate([repr, repr_p2])

            batch_cluster_repr.append(repr)

        All_cluster_repr.append(batch_cluster_repr)

    adata.uns['Cluster_repr'] = np.array(All_cluster_repr)

    # Normalization
    adata.uns['Cluster_repr'] = np.concatenate(adata.uns['Cluster_repr'])
    adata.uns['Cluster_repr'] = (adata.uns['Cluster_repr'] - adata.uns['Cluster_repr'].mean(axis=0)) / adata.uns['Cluster_repr'].std(axis=0)
    adata.uns['Cluster_repr'] = np.reshape(adata.uns['Cluster_repr'], (len(sorted_batch), len(np.unique(adata.obs['louvain'])), adata.uns['Cluster_repr'].shape[1]))

    return adata



class stClinic_Prediction_Model(nn.Module):
    def __init__(self, hidden_dims_pred, stat_dim):
        super(stClinic_Prediction_Model, self).__init__()

        # Set attention-based prediction network structure
        [in_dim, out_dim] = hidden_dims_pred

        self.in_dim = in_dim

        self.a_list = nn.ParameterList( [nn.Parameter(torch.empty(size=(stat_dim*in_dim, 1))) for i in range(stat_dim)] )
        [ nn.init.xavier_uniform_(a.data, gain=1.414) for a in self.a_list ]

        self.leakyrelu = nn.LeakyReLU(0.2)

        self.softmax = nn.Softmax(dim=1)

        self.fc = nn.Linear(in_dim, out_dim)

    def forward(self, data, pred_type):

        repr = data.x.transpose(1, 2).contiguous()

        w_list = [ torch.matmul(repr.view(repr.shape[0], -1), a.cuda()) for a in self.a_list ]
        e = self.leakyrelu( torch.cat(w_list, 1) )
        lambda_e = F.softmax(e)

        h1 = (lambda_e.unsqueeze(1).repeat(1, self.in_dim, 1).permute(0, 2, 1) * repr).sum(dim=1)

        h1 = self.softmax(h1)

        h2 = self.fc(h1)

        if pred_type=='survival':
            h2 = F.softplus(h2)
        elif pred_type=='grading':
            h2 = F.sigmoid(h2)

        return h2

    def COX_loss(self, surv_time, censor, hazard_pred):

        current_batch_len = len(surv_time)
        R_mat = np.zeros([current_batch_len, current_batch_len], dtype=int)

        for i in range(current_batch_len):
            for j in range(current_batch_len):
                R_mat[i, j] = (surv_time[j] >= surv_time[i])

        R_mat = torch.FloatTensor(R_mat).cuda()
        theta = hazard_pred.reshape(-1)
        exp_theta = torch.exp(theta)

        loss_cox = (-1) * torch.mean((theta - torch.log(torch.sum(exp_theta * R_mat, dim=1))) * censor)

        return loss_cox

    def load_model(self, path):
        """
        Load trained model parameters dictionary.
        Parameters
        ----------
        path
            file path that stores the model parameters
        """
        pretrained_dict = torch.load(path, map_location=lambda storage, loc: storage)
        model_dict = self.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if k in model_dict}
        model_dict.update(pretrained_dict) 
        self.load_state_dict(model_dict)



def train_Prediction_Model(adata, hidden_dims_pred=[1],
                           pred_type='survival',
                           n_epochs=100, lr=0.014, batch_name='batch_name',
                           gradient_clipping=5, weight_decay=0.00001, key_add='Cluster_importance',
                           random_seed=666, device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu'),
                           output_dir='./'):

    seed = random_seed
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)

    def extract_number(s):
        return int(re.findall(r'\d+', s)[0])
    sorted_batch = sorted( np.unique(adata.obs[batch_name]), key=extract_number)

    data_list = []
    for idx, bid in enumerate(sorted_batch):

        if pred_type=='survival':
            data = Data(x=torch.FloatTensor(adata.uns['Cluster_repr'][idx]),
                        y=torch.FloatTensor(adata.uns[pred_type][idx]),
                        graph_idx=bid)
        elif pred_type=='grading':
            data = Data(x=torch.FloatTensor(adata.uns['Cluster_repr'][idx]),
                        y=torch.FloatTensor([adata.uns[pred_type][idx]]),
                        graph_idx=bid)

        data = data.to(device)

        data_list.append(data)

    loader = DenseDataLoader(data_list, batch_size=8, shuffle=True)

    os.makedirs('/home/zuocm/Share_data/xiajunjie/stClinic/Prediction/checkpoint/', exist_ok=True)
    early_stopping = EarlyStopping(patience=10, checkpoint_file='/home/zuocm/Share_data/xiajunjie/stClinic/Prediction/checkpoint/StatAttention_UMAP_model.pt')

    model = stClinic_Prediction_Model(hidden_dims_pred=[data.x.shape[0], hidden_dims_pred[0]],
                                      stat_dim=data.x.shape[1]
                                     ).to(device)

    optimizer_pred = optim.Adam(filter(lambda p: p.requires_grad, model.parameters()), lr=lr, weight_decay=weight_decay)

    with trange(n_epochs, total=n_epochs, desc='Epochs') as tq:
        for epoch in tq:

            model.train()
            loss_total = 0

            for data in loader:

                optimizer_pred.zero_grad()

                if pred_type=='survival':
                    out = model(data, pred_type)
                    loss = model.COX_loss(data.y[:,1], data.y[:,0], out)
                elif pred_type=='grading':
                    out = model(data, pred_type)
                    loss = F.binary_cross_entropy(out, data.y)

                loss.backward()
                nn.utils.clip_grad_norm_(filter(lambda p: p.requires_grad, model.parameters()), gradient_clipping)
                optimizer_pred.step()

                loss_total += loss

            epoch_info = 'loss={:.5f}'.format(loss_total)
            tq.set_postfix_str(epoch_info)

            early_stopping(loss_total.cpu().detach().numpy(), model)
            if early_stopping.early_stop:
                print('EarlyStopping: run {} epoch'.format(epoch+1))
                break

    model.eval()
    model.zero_grad()

    adata.uns[key_add] = model.fc.weight.data.detach().cpu().numpy()[0]

    test_loader = DenseDataLoader(data_list, batch_size=1, shuffle=False)
    out_total = np.array([])

    for data in test_loader:
        if pred_type=='survival':
            out = model(data, pred_type)
            out_total = np.concatenate((out_total, out.detach().cpu().numpy().reshape(-1)))
        elif pred_type=='grading':
            out = model(data, pred_type)
            pred_grading = np.array([1]) if out>=0.5 else np.array([0])
            out_total = np.concatenate((out_total, pred_grading))

    if pred_type=='survival':
        print('# ============ Prognosis prediction ============= #')

        cindex_test = CIndex_lifeline(out_total, adata.uns[pred_type][:,0], adata.uns[pred_type][:,1])
        pvalue_test = cox_log_rank(out_total, adata.uns[pred_type][:,0], adata.uns[pred_type][:,1])
        surv_acc_test = accuracy_cox(out_total, adata.uns[pred_type][:,0])
        print('CIndex: {:.3f}, Pval: {:.3f}, Acc_COX: {:.3f}'.format(cindex_test, pvalue_test, surv_acc_test))

        KM_plot(out_total, adata.uns[pred_type][:,0], adata.uns[pred_type][:,1], output_dir=output_dir)

    elif pred_type=='grading':
        print('# ============ Grading prediction ============= #')

        grading_acc = np.mean(out_total == adata.uns['grading'])
        print('Grading Acc: {:.3f}'.format(grading_acc))

        CM_plot(out_total, adata.uns['grading'], output_dir=output_dir)
        ROC_plot(out_total, adata.uns['grading'], output_dir=output_dir)

    return adata



def train_Cross_Validation_Model(adata, hidden_dims_pred=[1],
                                 pred_type='grading',
                                 n_epochs=100, lr=0.0647, batch_name='batch_name',
                                 gradient_clipping=5, weight_decay=0.00001,
                                 random_seed=666, device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu'),
                                 output_dir='./'):

    seed = random_seed
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)

    def extract_number(s):
        return int(re.findall(r'\d+', s)[0])
    sorted_batch = sorted( np.unique(adata.obs[batch_name]), key=extract_number)

    data_list = []
    for idx, bid in enumerate(sorted_batch):

        if pred_type=='survival':
            data = Data(x=torch.FloatTensor(adata.uns['Cluster_repr'][idx]),
                        y=torch.FloatTensor(adata.uns[pred_type][idx]),
                        graph_idx=bid)
        elif pred_type=='grading':
            data = Data(x=torch.FloatTensor(adata.uns['Cluster_repr'][idx]),
                        y=torch.FloatTensor([adata.uns[pred_type][idx]]),
                        graph_idx=bid)

        data = data.to(device)

        data_list.append(data)

    if pred_type=='survival':

        # 7-fold cross-validation for prognosis prediction task

        c_index_set, acc_set = [], []

        for num_fold in range(7):

            train_idx = np.ones([len(data_list)], dtype=int)
            train_idx[ np.random.choice( np.arange(len(data_list)), int(len(data_list) / 7)+1, replace=False) ] = 0
            train_idx = train_idx==1

            train_data_list = [td for td, idx in zip(data_list, train_idx) if idx]
            test_data_list = [td for td, idx in zip(data_list, train_idx) if ~idx]

            print(f'No. {num_fold+1} fold. Train data size: {len(train_data_list)}. Test data size: {len(test_data_list)}.')

            train_loader = DenseDataLoader(train_data_list, batch_size=8, shuffle=True)

            os.makedirs(f'/home/zuocm/Share_data/xiajunjie/stClinic/CrossValidation/checkpoint_CV{num_fold+1}/', exist_ok=True)
            early_stopping = EarlyStopping(patience=10, checkpoint_file=f'/home/zuocm/Share_data/xiajunjie/stClinic/CrossValidation/checkpoint_CV{num_fold+1}/Stat_AttPred_UMAP_model_CV{num_fold+1}.pt')

            model = stClinic_Prediction_Model(hidden_dims_pred=[data.x.shape[0], hidden_dims_pred[0]],
                                                stat_dim=data.x.shape[1]
                                                ).to(device)

            optimizer_pred = optim.Adam(filter(lambda p: p.requires_grad, model.parameters()), lr=lr, weight_decay=weight_decay)

            with trange(n_epochs, total=n_epochs, desc='Epochs') as tq:
                for epoch in tq:

                    model.train()
                    loss_train = 0

                    for train_data in train_loader:

                        optimizer_pred.zero_grad()

                        out = model(train_data, pred_type)
                        loss = model.COX_loss(train_data.y[:,1], train_data.y[:,0], out)

                        loss.backward()
                        nn.utils.clip_grad_norm_(filter(lambda p: p.requires_grad, model.parameters()), gradient_clipping)
                        optimizer_pred.step()

                        loss_train += loss

                    epoch_info = 'train loss={:.5f}'.format(loss_train)
                    tq.set_postfix_str(epoch_info)

                    early_stopping(loss_train.cpu().detach().numpy(), model)
                    if early_stopping.early_stop:
                        print('EarlyStopping: run {} epoch'.format(epoch+1))
                        break

            model.eval()
            model.zero_grad()

            test_loader = DenseDataLoader(test_data_list, batch_size=1, shuffle=False)

            out_total = np.array([])
            for test_data in test_loader:
                out = model(test_data, pred_type)
                out_total = np.concatenate((out_total, out.detach().cpu().numpy().reshape(-1)))
            
            cindex_test = CIndex_lifeline(out_total, adata.uns[pred_type][:,0][~train_idx], adata.uns[pred_type][:,1][~train_idx])
            surv_acc_test = accuracy_cox(out_total, adata.uns[pred_type][:,0][~train_idx])

            print('No. {:d} fold. COX: {:.3f}, ACC: {:.3f}'.format(num_fold+1, cindex_test, surv_acc_test))

            c_index_set.append(cindex_test)
            acc_set.append(surv_acc_test)

        c_index_avg = np.mean(c_index_set)
        acc_avg = np.mean(acc_set)

        print('7-fold cross-validation. COX: {:.3f}, ACC: {:.3f}'.format(c_index_avg, acc_avg))

        return c_index_avg, acc_avg

    elif pred_type=='grading':

        # Leava-one-out cross-validation for grading prediction task

        out_total = np.array([])

        for num_fold in range(len(data_list)):

            train_idx = np.ones([len(data_list)], dtype=int)
            train_idx[num_fold] = 0
            train_idx = train_idx==1

            train_data_list = [td for td, idx in zip(data_list, train_idx) if idx]
            test_data_list = [td for td, idx in zip(data_list, train_idx) if ~idx]

            print(f'No. {num_fold+1} fold. Train data size: {len(train_data_list)}. Test data size: {len(test_data_list)}.')

            train_loader = DenseDataLoader(train_data_list, batch_size=8, shuffle=True)

            os.makedirs(f'/home/zuocm/Share_data/xiajunjie/stClinic/CrossValidation/checkpoint_CV{num_fold+1}/', exist_ok=True)
            early_stopping = EarlyStopping(patience=10, checkpoint_file=f'/home/zuocm/Share_data/xiajunjie/stClinic/CrossValidation/checkpoint_CV{num_fold+1}/Stat_AttPred_UMAP_model_CV{num_fold+1}.pt')

            model = stClinic_Prediction_Model(hidden_dims_pred=[data.x.shape[0], hidden_dims_pred[0]],
                                              stat_dim=data.x.shape[1]
                                             ).to(device)

            optimizer_pred = optim.Adam(filter(lambda p: p.requires_grad, model.parameters()), lr=lr, weight_decay=weight_decay)

            with trange(n_epochs, total=n_epochs, desc='Epochs') as tq:
                for epoch in tq:

                    model.train()
                    loss_train = 0

                    for train_data in train_loader:

                        optimizer_pred.zero_grad()

                        out = model(train_data, pred_type)
                        loss = F.binary_cross_entropy(out, train_data.y)

                        loss.backward()
                        nn.utils.clip_grad_norm_(filter(lambda p: p.requires_grad, model.parameters()), gradient_clipping)
                        optimizer_pred.step()

                        loss_train += loss

                    epoch_info = 'train loss={:.5f}'.format(loss_train)
                    tq.set_postfix_str(epoch_info)

                    early_stopping(loss_train.cpu().detach().numpy(), model)
                    if early_stopping.early_stop:
                        print('EarlyStopping: run {} epoch'.format(epoch+1))
                        break

            model.eval()
            model.zero_grad()

            test_loader = DenseDataLoader(test_data_list, batch_size=1, shuffle=False)

            for test_data in test_loader:
                out = model(test_data, pred_type)
                pred_grading = np.array([1]) if out>=0.5 else np.array([0])

            out_total = np.concatenate((out_total, pred_grading))

        grading_acc = np.mean(out_total == adata.uns['grading'])

        print('Leave-one-out cross-validation. Grading Acc: {:.3f}'.format(grading_acc))

        ROC_plot(out_total, adata.uns['grading'], output_dir=output_dir)
        CM_plot(out_total, adata.uns['grading'], output_dir=output_dir)

        return grading_acc


import numpy as np
import pandas as pd
import itertools
import networkx as nx
import hnswlib
import scipy.linalg

from sklearn.neighbors import NearestNeighbors
from scipy.sparse import coo_matrix

def create_dictionary_mnn(adata, use_rep, batch_name, k = 5, save_on_disk = True, approx = True, iter_comb = None):

    cell_names = adata.obs_names

    batch_list = adata.obs[batch_name]
    datasets = []
    datasets_pcs = []
    cells = []
    for i in batch_list.unique():
        datasets.append(adata[batch_list == i])
        datasets_pcs.append(adata[batch_list == i].obsm[use_rep])
        cells.append(cell_names[batch_list == i])

    batch_name_df = pd.DataFrame(np.array(batch_list.unique()))
    mnns = dict()

    if iter_comb is None:
        iter_comb = list(itertools.combinations(range(len(cells)), 2))
    for comb in iter_comb:
        i = comb[0]
        j = comb[1]
        key_name1 = batch_name_df.loc[comb[0]].values[0] + "_" + batch_name_df.loc[comb[1]].values[0]
        mnns[key_name1] = {}

        new = list(cells[j])
        ref = list(cells[i])

        ds1 = adata[new].obsm[use_rep]
        ds2 = adata[ref].obsm[use_rep]

        match = mnn(ds1, ds2, knn=k, save_on_disk = save_on_disk, approx = approx)
        
        coordinates = list(match)

        sparse_matrix = coo_matrix((np.ones(len(coordinates)), (zip(*coordinates))), shape=(len(ref), len(new)))

        mnns[key_name1] = sparse_matrix
    return(mnns)

def nn_approx(ds1, ds2, knn=5):
    dim = ds2.shape[1]
    num_elements = ds2.shape[0]
    p = hnswlib.Index(space='l2', dim=dim)
    p.init_index(max_elements=num_elements, ef_construction=100, M = 16)
    p.set_ef(10)
    p.add_items(ds2)
    ind,  distances = p.knn_query(ds1, k=knn)
    match = set()
    for a, b in zip(range(ds1.shape[0]), ind):
        for b_i in b:
            match.add((a, b_i))
    return match

def nn(ds1, ds2, knn=5, metric_p=2):

    nn_ = NearestNeighbors(knn, p=metric_p)
    nn_.fit(ds2)
    ind = nn_.kneighbors(ds1, return_distance=False)

    match = set()
    for a, b in zip(range(ds1.shape[0]), ind):
        for b_i in b:
            match.add((a, b_i))

    return match

def mnn(ds1, ds2, knn = 5, save_on_disk = True, approx = True):
    if approx: 
        # Find nearest neighbors in first direction.
        # output KNN point for each point in ds1.  match1 is a set(): (points in names1, points in names2), the size of the set is ds1.shape[0]*knn
        match1 = nn_approx(ds1, ds2, knn=knn)
        # Find nearest neighbors in second direction.
        match2 = nn_approx(ds2, ds1, knn=knn)
    else:
        match1 = nn(ds1, ds2, knn=knn)
        match2 = nn(ds2, ds1, knn=knn)
    # Compute mutual nearest neighbors.
    mutual = match2 & set([ (b, a) for a, b in match1 ])

    return mutual

def inter_linked_graph(adj_list, section_ids, mnn_dict):

    if mnn_dict:
        
        # inter_linked_graph

        adj_concat = np.asarray(adj_list[0].todense())
        for batch_id in range(1,len(section_ids)):

            last_batch_len = adj_concat.shape[0]

            adj_concat = scipy.linalg.block_diag(adj_concat, np.asarray(adj_list[batch_id].todense()))

            aux_mat = []
            for aux_id in range(batch_id):

                key_name1 = section_ids[aux_id] + '_' + section_ids[batch_id]

                aux_mat.append(mnn_dict[key_name1].toarray())
            aux_mat = np.concatenate(aux_mat, axis=0)

            adj_concat[:last_batch_len, last_batch_len:] = aux_mat
            adj_concat[last_batch_len:, :last_batch_len] = aux_mat.T

    else:
        
        # graph without inter-slice edges

        adj_concat = np.asarray(adj_list[0].todense())
        for batch_id in range(1,len(section_ids)):
            adj_concat = scipy.linalg.block_diag(adj_concat, np.asarray(adj_list[batch_id].todense()))

    return adj_concat

def count_spatial_neighbors(Batch_list, section_ids, adj_concat):

    start_num = 0
    for batch_id in range(len(Batch_list)):

        spot_num = Batch_list[batch_id].n_obs

        print(f'\n------Counting {section_ids[batch_id]} spatial neighbors...')

        avg_neigh_num = np.mean(np.sum(adj_concat[start_num:start_num+spot_num,:]==1,axis=1)-1)
        print(f'{round(avg_neigh_num,4)} neighbors per spot on average.')

        avg_intra_neigh_num = np.mean(np.sum(adj_concat[start_num:start_num+spot_num,start_num:start_num+spot_num]==1,axis=1)-1)
        print(f'{round(avg_intra_neigh_num,4)} intra-neighbors per spot on average.')

        avg_inter_neigh_num = avg_neigh_num - avg_intra_neigh_num
        print(f'{round(avg_inter_neigh_num,4)} inter-neighbors per spot on average.')

        start_num += spot_num


from setuptools import setup, find_packages

setup(
    name = 'stClinic',
    version = '0.0.1',
    keywords='TME',
    description = 'stClinic for dissecting clinically relevant niches by integrating spatial multi-slice multi-omics data',
    license = 'MIT License',
    url = 'https://github.com/cmzuo11/stClinic',
    author = 'cmzuo',
    author_email = 'cmzuo13@163.com',
    packages = find_packages(),
    include_package_data = True,
    platforms = 'any',
    install_requires = [
        'anndata==0.9.2',
        'numpy==1.22.3',
        'igraph==0.10.8',
        'torch==1.12.0',
        'torch-cluster==1.6.1',
        'torch-geometric==2.3.0',
        'torch-scatter==2.1.1',
        'torch-sparse==0.6.17',
        'torch-spline-conv==1.2.2',
        'torchaudio==0.12.0',
        'torchvision==0.13.0',
        'tqdm==4.65.0',
        'umap-learn==0.5.3',
        'h5py==3.9.0',
        'hnswlib==0.5.1',
        'rpy2==3.5.1',
        'scanpy==1.9.3',
        'scikit-learn==1.3.0',
        'scikit-misc==0.2.0',
        'scipy==1.10.1',
        'seaborn==0.11.2',
        'lifelines==0.27.8',
        'llvmlite==0.40.1',
        'louvain==0.8.1',
        'matplotlib==3.7.2',
        'setuptools==68.0.0',
        'networkx==3.1',
        'psutil==5.9.5',
        'pycparser==2.21',
        'pyg-lib==0.2.0+pt112cu113',
        'pynndescent==0.5.10',
        'pyparsing==3.0.9',
        'PySocks==1.7.1',
        'python-dateutil==2.8.2',
        'python-igraph==0.10.8',
        'pip==23.1.2',
        ],
        )
